import os

from shinya.bd.extension_data import ExtensionData
from shinya.common.info_dict import InfoDict
from shinya.common.io import unpack_bytes, pack_bytes


class MPLSHeader(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        self["TypeIndicator"] = data[0:4].decode("utf-8")
        self["VersionNumber"] = data[4:8].decode("utf-8")
        self["PlayListStartAddress"] = unpack_bytes(data, 8, 4)
        self["PlayListMarkStartAddress"] = unpack_bytes(data, 12, 4)
        self["ExtensionDataStartAddress"] = unpack_bytes(data, 16, 4)
        self["reserved1"] = data[20:40]

        appinfo_display_size = unpack_bytes(data, 40, 4)
        playlist_display_size = unpack_bytes(data, self["PlayListStartAddress"], 4)
        playlist_mark_display_size = unpack_bytes(data, self["PlayListMarkStartAddress"], 4)

        if self["ExtensionDataStartAddress"]:
            extension_display_size = unpack_bytes(data, self["ExtensionDataStartAddress"], 4)

        assert appinfo_display_size == 14
        assert self["PlayListStartAddress"] == 58
        assert (self["PlayListStartAddress"] + playlist_display_size + 4 == self["PlayListMarkStartAddress"])
        if self["ExtensionDataStartAddress"]:
            assert (self["PlayListMarkStartAddress"] + playlist_mark_display_size + 4 == self[
                "ExtensionDataStartAddress"])
            assert self["ExtensionDataStartAddress"] + extension_display_size + 4 == len(data)
        else:
            assert self["PlayListMarkStartAddress"] + playlist_mark_display_size + 4 == len(data)

        self["AppInfoPlayList"] = AppInfoPlayList.from_bytes(data[40: 40 + appinfo_display_size + 4])
        self["PlayList"] = PlayList.from_bytes(
            data[self["PlayListStartAddress"]: self["PlayListStartAddress"] + playlist_display_size + 4])
        self["PlayListMark"] = PlayListMark.from_bytes(
            data[self["PlayListMarkStartAddress"]: self["PlayListMarkStartAddress"] + playlist_mark_display_size + 4])
        if self["ExtensionDataStartAddress"]:
            self["ExtensionData"] = ExtensionData.from_bytes(
                data[self["ExtensionDataStartAddress"]: self["ExtensionDataStartAddress"] + extension_display_size + 4],
                self["ExtensionDataStartAddress"])

        assert data == self.to_bytes()
        return self

    def update_addresses(self, offset=0):
        playlist_display_size = self["PlayList"].calculate_display_size()
        playlist_mark_display_size = self["PlayListMark"].calculate_display_size()
        self["PlayListMarkStartAddress"] = self["PlayListStartAddress"] + playlist_display_size + 4
        if self["ExtensionDataStartAddress"]:
            self["ExtensionDataStartAddress"] = self["PlayListMarkStartAddress"] + playlist_mark_display_size + 4

    def check_constraints(self):
        appinfo_display_size = self["AppInfoPlayList"].calculate_display_size()
        playlist_display_size = self["PlayList"].calculate_display_size()
        playlist_mark_display_size = self["PlayListMark"].calculate_display_size()

        assert appinfo_display_size == 14
        assert self["PlayListStartAddress"] == 58
        assert (self["PlayListStartAddress"] + playlist_display_size + 4 == self["PlayListMarkStartAddress"])
        if self["ExtensionDataStartAddress"]:
            assert (self["PlayListMarkStartAddress"] + playlist_mark_display_size + 4 == self[
                "ExtensionDataStartAddress"])

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += self["TypeIndicator"].encode("utf-8")
        data += self["VersionNumber"].encode("utf-8")
        data += pack_bytes(self["PlayListStartAddress"], 4)
        data += pack_bytes(self["PlayListMarkStartAddress"], 4)
        data += pack_bytes(self["ExtensionDataStartAddress"], 4)
        data += self["reserved1"]

        data += self["AppInfoPlayList"].to_bytes()
        data += self["PlayList"].to_bytes()
        data += self["PlayListMark"].to_bytes()
        if self["ExtensionDataStartAddress"]:
            data += self["ExtensionData"].to_bytes(self["ExtensionDataStartAddress"])
        return data


class AppInfoPlayList(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        assert len(data) == 18

        self = cls()
        self["Length"] = unpack_bytes(data, 0, 4)
        self["reserved1"] = unpack_bytes(data, 4, 1)
        self["PlaybackType"] = unpack_bytes(data, 5, 1)
        if self["PlaybackType"] in [2, 3]:
            self["PlaybackCount"] = unpack_bytes(data, 6, 2)
        else:
            self["reserved2"] = unpack_bytes(data, 6, 2)

        self["UOMaskTable"] = UOMaskTable.from_bytes(data[8:16])
        flags = unpack_bytes(data, 16, 2)

        self["RandomAccessFlag"] = flags >> 15 & 1
        self["AudioMixFlag"] = flags >> 14 & 1
        self["LosslessBypassFlag"] = flags >> 13 & 1
        self["MVCBaseViewRFlag"] = flags >> 12 & 1
        self["SDRConversionNotificationFlag"] = flags >> 11 & 1
        self["reserved3"] = flags % 2 ** 11

        return self

    def calculate_display_size(self):
        return 14

    def to_bytes(self):

        self.check_constraints()

        flags = (
                (self["RandomAccessFlag"] << 15)
                + (self["AudioMixFlag"] << 14)
                + (self["LosslessBypassFlag"] << 13)
                + (self["MVCBaseViewRFlag"] << 12)
                + (self["SDRConversionNotificationFlag"] << 11)
                + self["reserved3"]
        )

        data = b""
        data += pack_bytes(self["Length"], 4)
        data += pack_bytes(self["reserved1"], 1)
        data += pack_bytes(self["PlaybackType"], 1)

        if self["PlaybackType"] in [2, 3]:
            data += pack_bytes(self["PlaybackCount"], 2)
        else:
            data += pack_bytes(self["reserved2"], 2)

        data += self["UOMaskTable"].to_bytes()
        data += pack_bytes(flags, 2)

        return data


class UOMaskTable(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        assert len(data) == 8

        self = cls()
        uo_mask_table = unpack_bytes(data, 0, 8)
        self["MenuCall"] = uo_mask_table >> 63 & 1
        self["TitleSearch"] = uo_mask_table >> 62 & 1
        self["ChapterSearch"] = uo_mask_table >> 61 & 1
        self["TimeSearch"] = uo_mask_table >> 60 & 1
        self["SkipToNextPoint"] = uo_mask_table >> 59 & 1
        self["SkipToPrevPoint"] = uo_mask_table >> 58 & 1
        self["reserved1"] = uo_mask_table >> 57 & 1
        self["Stop"] = uo_mask_table >> 56 & 1
        self["PauseOn"] = uo_mask_table >> 55 & 1
        self["reserved2"] = uo_mask_table >> 54 & 1
        self["StillOff"] = uo_mask_table >> 53 & 1
        self["ForwardPlay"] = uo_mask_table >> 52 & 1
        self["BackwardPlay"] = uo_mask_table >> 51 & 1
        self["Resume"] = uo_mask_table >> 50 & 1
        self["MoveUpSelectedButton"] = uo_mask_table >> 49 & 1
        self["MoveDownSelectedButton"] = uo_mask_table >> 48 & 1
        self["MoveLeftSelectedButton"] = uo_mask_table >> 47 & 1
        self["MoveRightSelectedButton"] = uo_mask_table >> 46 & 1
        self["SelectButton"] = uo_mask_table >> 45 & 1
        self["ActivateButton"] = uo_mask_table >> 44 & 1
        self["SelectAndActivateButton"] = uo_mask_table >> 43 & 1
        self["PrimaryAudioStreamNumberChange"] = uo_mask_table >> 42 & 1
        self["reserved3"] = uo_mask_table >> 41 & 1
        self["AngleNumberChange"] = uo_mask_table >> 40 & 1
        self["PopupOn"] = uo_mask_table >> 39 & 1
        self["PopupOff"] = uo_mask_table >> 38 & 1
        self["PrimaryPGEnableDisable"] = uo_mask_table >> 37 & 1
        self["PrimaryPGStreamNumberChange"] = uo_mask_table >> 36 & 1
        self["SecondaryVideoEnableDisable"] = uo_mask_table >> 35 & 1
        self["SecondaryVideoStreamNumberChange"] = uo_mask_table >> 34 & 1
        self["SecondaryAudioEnableDisable"] = uo_mask_table >> 33 & 1
        self["SecondaryAudioStreamNumberChange"] = uo_mask_table >> 32 & 1
        self["reserved4"] = uo_mask_table >> 31 & 1
        self["SecondaryPGStreamNumberChange"] = uo_mask_table >> 30 & 1
        self["reserved5"] = uo_mask_table % 2 ** 30

        return self

    def to_bytes(self):
        self.check_constraints()

        uo_mask_table = (
                (self["MenuCall"] << 63)
                + (self["TitleSearch"] << 62)
                + (self["ChapterSearch"] << 61)
                + (self["TimeSearch"] << 60)
                + (self["SkipToNextPoint"] << 59)
                + (self["SkipToPrevPoint"] << 58)
                + (self["reserved1"] << 57)
                + (self["Stop"] << 56)
                + (self["PauseOn"] << 55)
                + (self["reserved2"] << 54)
                + (self["StillOff"] << 53)
                + (self["ForwardPlay"] << 52)
                + (self["BackwardPlay"] << 51)
                + (self["Resume"] << 50)
                + (self["MoveUpSelectedButton"] << 49)
                + (self["MoveDownSelectedButton"] << 48)
                + (self["MoveLeftSelectedButton"] << 47)
                + (self["MoveRightSelectedButton"] << 46)
                + (self["SelectButton"] << 45)
                + (self["ActivateButton"] << 44)
                + (self["SelectAndActivateButton"] << 43)
                + (self["PrimaryAudioStreamNumberChange"] << 42)
                + (self["reserved3"] << 41)
                + (self["AngleNumberChange"] << 40)
                + (self["PopupOn"] << 39)
                + (self["PopupOff"] << 38)
                + (self["PrimaryPGEnableDisable"] << 37)
                + (self["PrimaryPGStreamNumberChange"] << 36)
                + (self["SecondaryVideoEnableDisable"] << 35)
                + (self["SecondaryVideoStreamNumberChange"] << 34)
                + (self["SecondaryAudioEnableDisable"] << 33)
                + (self["SecondaryAudioStreamNumberChange"] << 32)
                + (self["reserved4"] << 31)
                + (self["SecondaryPGStreamNumberChange"] << 30)
                + self["reserved5"]
        )

        return pack_bytes(uo_mask_table, 8)


class PlayList(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):

        self = cls()
        self["Length"] = unpack_bytes(data, 0, 4)
        self["reserved1"] = unpack_bytes(data, 4, 2)
        self["NumberOfPlayItems"] = unpack_bytes(data, 6, 2)
        self["NumberOfSubPaths"] = unpack_bytes(data, 8, 2)
        self["PlayItems"] = []
        self["SubPaths"] = []

        read_index = 10
        for i in range(self["NumberOfPlayItems"]):
            item_length = unpack_bytes(data, read_index, 2)
            self["PlayItems"].append(
                PlayItem.from_bytes(data[read_index: read_index + item_length + 2])
            )
            read_index += item_length + 2

        for i in range(self["NumberOfSubPaths"]):
            item_length = unpack_bytes(data, read_index, 4)
            self["SubPaths"].append(
                SubPath.from_bytes(data[read_index: read_index + item_length + 4])
            )
            read_index += item_length + 4

        return self

    def calculate_display_size(self):
        real_length = 10
        for i in self["PlayItems"]:
            real_length += i.calculate_display_size() + 2
        for i in self["SubPaths"]:
            real_length += i.calculate_display_size() + 4
        return real_length - 4

    def update_counts(self):
        self["NumberOfPlayItems"] = len(self["PlayItems"])
        self["NumberOfSubPaths"] = len(self["SubPaths"])

    def check_constraints(self):
        super().check_constraints()
        assert self["NumberOfPlayItems"] == len(self["PlayItems"])
        assert self["NumberOfSubPaths"] == len(self["SubPaths"])

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["Length"], 4)
        data += pack_bytes(self["reserved1"], 2)
        data += pack_bytes(self["NumberOfPlayItems"], 2)
        data += pack_bytes(self["NumberOfSubPaths"], 2)
        for i in self["PlayItems"]:
            data += i.to_bytes()
        for i in self["SubPaths"]:
            data += i.to_bytes()
        return data


class PlayItem(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):

        self = cls()
        self["Length"] = unpack_bytes(data, 0, 2)
        self["ClipInformationFileName"] = data[2:7].decode("utf-8")
        self["ClipCodecIdentifier"] = data[7:11].decode("utf-8")
        flags = unpack_bytes(data, 11, 2)
        self["reserved1"], flags = divmod(flags, 2 ** 5)
        self["IsMultiAngle"], self["ConnectionCondition"] = divmod(flags, 2 ** 4)
        self["RefToSTCID"] = unpack_bytes(data, 13, 1)
        self["INTime"] = unpack_bytes(data, 14, 4)
        self["OUTTime"] = unpack_bytes(data, 18, 4)
        self["UOMaskTable"] = UOMaskTable.from_bytes(data[22:30])
        flags = unpack_bytes(data, 30, 1)
        self["PlayItemRandomAccessFlag"], self["reserved2"] = divmod(flags, 2 ** 7)
        self["StillMode"] = unpack_bytes(data, 31, 1)
        if self["StillMode"] == 1:
            self["StillTime"] = unpack_bytes(data, 32, 2)
        else:
            self["reserved3"] = unpack_bytes(data, 32, 2)

        read_index = 34

        if self["IsMultiAngle"]:
            self["NumberOfAngles"] = unpack_bytes(data, 34, 1)
            flags = unpack_bytes(data, 35, 1)
            self["reserved4"] = flags // 2 ** 2
            flags %= 2 ** 2
            self["IsDifferentAudios"] = flags // 2
            self["IsSeamlessAngleChange"] = flags % 2
            self["Angles"] = []
            read_index += 2
            for i in range(self["NumberOfAngles"] - 1):
                self["Angles"].append(
                    MultiClipEntry.from_bytes(data[read_index: read_index + 10])
                )
                read_index += 10

        self["STNTable"] = STNTable.from_bytes(data[read_index:])

        return self

    def calculate_display_size(self):
        real_length = 34
        if self["IsMultiAngle"]:
            real_length += 2
            real_length += len(self["Angles"]) * 10
        real_length += self["STNTable"].calculate_display_size() + 2
        return real_length - 2

    def update_counts(self):
        if self["IsMultiAngle"]:
            self["NumberOfAngles"] = len(self["Angles"]) + 1

    def check_constraints(self):
        super().check_constraints()
        if self["IsMultiAngle"]:
            assert self["NumberOfAngles"] - 1 == len(self["Angles"])

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["Length"], 2)
        data += self["ClipInformationFileName"].encode("utf-8")
        data += self["ClipCodecIdentifier"].encode("utf-8")
        flags = (
                (self["reserved1"] << 5)
                + (self["IsMultiAngle"] << 4)
                + self["ConnectionCondition"]
        )
        data += pack_bytes(flags, 2)
        data += pack_bytes(self["RefToSTCID"], 1)
        data += pack_bytes(self["INTime"], 4)
        data += pack_bytes(self["OUTTime"], 4)
        data += self["UOMaskTable"].to_bytes()
        data += pack_bytes(
            (self["PlayItemRandomAccessFlag"] << 7) + self["reserved2"], 1
        )
        data += pack_bytes(self["StillMode"], 1)
        if self["StillMode"] == 1:
            data += pack_bytes(self["StillTime"], 2)
        else:
            data += pack_bytes(self["reserved3"], 2)

        if self["IsMultiAngle"]:
            data += pack_bytes(self["NumberOfAngles"], 1)
            data += pack_bytes(
                (self["reserved4"] << 2)
                + (self["IsDifferentAudios"] << 1)
                + self["IsSeamlessAngleChange"],
                1
            )
            for i in self["Angles"]:
                data += i.to_bytes()

        data += self["STNTable"].to_bytes()

        return data


class STNTable(InfoDict):
    stream_names = [
        "PrimaryVideoStreamEntries",
        "PrimaryAudioStreamEntries",
        "PrimaryPGStreamEntries",
        "PrimaryIGStreamEntries",
        "SecondaryAudioStreamEntries",
        "SecondaryVideoStreamEntries",
        "SecondaryPGStreamEntries",
        "DVStreamEntries",
    ]

    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):

        self = cls()
        self["Length"] = unpack_bytes(data, 0, 2)
        if self["Length"] != 0:
            self["reserved1"] = unpack_bytes(data, 2, 2)

            read_index = 4
            for name in self.stream_names:
                self[f"NumberOf{name}"] = unpack_bytes(data, read_index, 1)
                read_index += 1

            self["reserved2"] = unpack_bytes(data, 12, 4)
            read_index = 16

            for name in self.stream_names:
                self[name] = []
                for i in range(self[f"NumberOf{name}"]):
                    info_pair = InfoDict()

                    stream_entry_length = unpack_bytes(data, read_index, 1)
                    info_pair["StreamEntry"] = StreamEntry.from_bytes(
                        data[read_index: read_index + stream_entry_length + 1]
                    )
                    read_index += stream_entry_length + 1

                    stream_attr_length = unpack_bytes(data, read_index, 1)
                    info_pair["StreamAttributes"] = StreamAttributes.from_bytes(
                        data[read_index: read_index + stream_attr_length + 1]
                    )
                    read_index += stream_attr_length + 1

                    self[name].append(info_pair)

        return self

    def calculate_display_size(self):
        if self["Length"] != 0:
            real_length = 16
            for name in self.stream_names:
                for i in self[name]:
                    real_length += (
                            i["StreamEntry"].calculate_display_size() + 1
                            + i["StreamAttributes"].calculate_display_size() + 1
                    )
            return real_length - 2
        else:
            return 0

    def update_counts(self):
        if self["Length"] != 0:
            for name in self.stream_names:
                self[f"NumberOf{name}"] = len(self[name])

    def check_constraints(self):
        super().check_constraints()
        if self["Length"] != 0:
            for name in self.stream_names:
                assert self[f"NumberOf{name}"] == len(self[name])

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["Length"], 2)
        if self["Length"] != 0:
            data += pack_bytes(self["reserved1"], 2)
            for name in self.stream_names:
                data += pack_bytes(self[f"NumberOf{name}"], 1)
            data += pack_bytes(self["reserved2"], 4)
            for name in self.stream_names:
                for i in self[name]:
                    data += i["StreamEntry"].to_bytes()
                    data += i["StreamAttributes"].to_bytes()
        return data


class StreamEntry(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        self["Length"] = unpack_bytes(data, 0, 1)
        if self["Length"] != 0:
            self["StreamType"] = unpack_bytes(data, 1, 1)
            if self["StreamType"] == 1:
                self["RefToStreamPID"] = unpack_bytes(data, 2, 2)
            elif self["StreamType"] == 2:
                self["RefToSubPathID"] = unpack_bytes(data, 2, 1)
                self["RefToSubClipID"] = unpack_bytes(data, 3, 1)
                self["RefToStreamPID"] = unpack_bytes(data, 4, 2)
            elif self["StreamType"] in [3, 4]:
                self["RefToSubPathID"] = unpack_bytes(data, 2, 1)
                self["RefToStreamPID"] = unpack_bytes(data, 3, 2)
            else:
                assert False

        return self

    def calculate_display_size(self):
        if self["Length"] != 0:
            return 9
        else:
            return 0

    # def check_constraints(self):
    #     if self["Length"] != 0:
    #         self["Length"] + 1 == 10
    #
    #         if self["StreamType"] == 1:
    #             assert self["Length"] + 1 == 4
    #         elif self["StreamType"] == 2:
    #             assert self["Length"] + 1 == 6
    #         elif self["StreamType"] in [3, 4]:
    #             assert self["Length"] + 1 == 5
    #         else:
    #             assert False

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["Length"], 1)
        if self["Length"] != 0:
            data += pack_bytes(self["StreamType"], 1)
            if self["StreamType"] == 1:
                data += pack_bytes(self["RefToStreamPID"], 2)
                data += b"\x00\x00\x00\x00\x00\x00"
            elif self["StreamType"] == 2:
                data += pack_bytes(self["RefToSubPathID"], 1)
                data += pack_bytes(self["RefToSubClipID"], 1)
                data += pack_bytes(self["RefToStreamPID"], 2)
                data += b"\x00\x00\x00\x00"
            elif self["StreamType"] in [3, 4]:
                data += pack_bytes(self["RefToSubPathID"], 1)
                data += pack_bytes(self["RefToStreamPID"], 2)
                data += b"\x00\x00\x00\x00\x00"
            return data


class StreamAttributes(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):

        self = cls()
        self["Length"] = unpack_bytes(data, 0, 1)
        if self["Length"] != 0:
            self["StreamCodingType"] = unpack_bytes(data, 1, 1)
            if self["StreamCodingType"] in [0x01, 0x02, 0x1B, 0xEA]:
                tmp = unpack_bytes(data, 2, 1)
                self["VideoFormat"], self["FrameRate"] = divmod(tmp, 2 ** 4)

            elif self["StreamCodingType"] == 0x24:
                tmp = unpack_bytes(data, 2, 1)
                self["VideoFormat"], self["FrameRate"] = divmod(tmp, 2 ** 4)
                tmp = unpack_bytes(data, 3, 1)
                self["DynamicRangeType"], self["ColorSpace"] = divmod(tmp, 2 ** 4)
                tmp = unpack_bytes(data, 4, 1)
                self["CRFlag"] = tmp >> 7 & 1
                self["HDRPlusFlag"] = tmp >> 6 & 1

            elif self["StreamCodingType"] in [0x03, 0x04, 0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0xA1, 0xA2]:
                tmp = unpack_bytes(data, 2, 1)
                self["AudioFormat"], self["SampleRate"] = divmod(tmp, 2 ** 4)
                self["LanguageCode"] = data[3:6].decode("utf-8")

            elif self["StreamCodingType"] in [0x90, 0x91]:
                self["LanguageCode"] = data[2:5].decode("utf-8")

            elif self["StreamCodingType"] in [0x92]:
                self["CharacterCode"] = unpack_bytes(data, 2, 1)
                self["LanguageCode"] = data[3:6].decode("utf-8")

            else:
                assert False

        return self

    def calculate_display_size(self):
        if self["Length"] != 0:
            return 5
        else:
            return 0

    #     def check_constraints(self):
    #         if self["Length"] != 0:
    #             assert self["Length"] + 1 == 6

    #             if self["StreamCodingType"] in [0x01, 0x02, 0x1B, 0xEA]:
    #                 assert self["Length"] + 1 == 3

    #             elif self["StreamCodingType"] == 0x24:
    #                 assert self["Length"] + 1 == 5

    #             elif self["StreamCodingType"] in [
    #                 0x03,
    #                 0x04,
    #                 0x80,
    #                 0x81,
    #                 0x82,
    #                 0x83,
    #                 0x84,
    #                 0x85,
    #                 0x86,
    #                 0xA1,
    #                 0xA2,
    #             ]:
    #                 assert self["Length"] + 1 == 6

    #             elif self["StreamCodingType"] in [0x90, 0x91]:
    #                 assert self["Length"] + 1 == 5

    #             elif self["StreamCodingType"] in [0x92]:
    #                 assert self["Length"] + 1 == 6

    #             else:
    #                 assert False

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["Length"], 1)
        if self["Length"] != 0:
            data += pack_bytes(self["StreamCodingType"], 1)
            if self["StreamCodingType"] in [0x01, 0x02, 0x1B, 0xEA]:
                data += pack_bytes(
                    (self["VideoFormat"] << 4) + self["FrameRate"], 1
                )
                data += b"\x00\x00\x00"

            elif self["StreamCodingType"] == 0x24:
                data += pack_bytes(
                    (self["VideoFormat"] << 4) + self["FrameRate"], 1
                )
                data += pack_bytes(
                    (self["DynamicRangeType"] << 4) + self["ColorSpace"], 1
                )
                data += pack_bytes(
                    (self["CRFlag"] << 7) + (self["HDRPlusFlag"] << 6), 1
                )
                data += b"\x00"

            elif self["StreamCodingType"] in [0x03, 0x04, 0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0xA1, 0xA2]:
                data += pack_bytes(
                    (self["AudioFormat"] << 4) + self["SampleRate"], 1
                )
                data += self["LanguageCode"].encode("utf-8")

            elif self["StreamCodingType"] in [0x90, 0x91]:
                data += self["LanguageCode"].encode("utf-8")
                data += b"\x00"

            elif self["StreamCodingType"] in [0x92]:
                data += pack_bytes(self["CharacterCode"], 1)
                data += self["LanguageCode"].encode("utf-8")

        return data


class SubPath(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):

        self = cls()
        self["Length"] = unpack_bytes(data, 0, 4)
        self["reserved1"] = unpack_bytes(data, 4, 1)
        self["SubPathType"] = unpack_bytes(data, 5, 1)
        flags = unpack_bytes(data, 6, 2)
        self["reserved2"] = flags // 2
        self["IsRepeatSubPath"] = flags % 2
        self["reserved3"] = unpack_bytes(data, 8, 1)
        self["NumberOfSubPlayItems"] = unpack_bytes(data, 9, 1)
        self["SubPlayItems"] = []
        read_index = 10
        for i in range(self["NumberOfSubPlayItems"]):
            item_length = unpack_bytes(data, read_index, 2)
            self["SubPlayItems"].append(SubPlayItem.from_bytes(data[read_index: read_index + item_length + 2]))
            read_index += item_length + 2

        return self

    def calculate_display_size(self):
        real_length = 10
        for i in self["SubPlayItems"]:
            real_length += i.calculate_display_size() + 2
        return real_length - 4

    def update_counts(self):
        self["NumberOfSubPlayItems"] = len(self["SubPlayItems"])

    def check_constraints(self):
        super().check_constraints()
        assert self["NumberOfSubPlayItems"] == len(self["SubPlayItems"])

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["Length"], 4)
        data += pack_bytes(self["reserved1"], 1)
        data += pack_bytes(self["SubPathType"], 1)
        data += pack_bytes(self["reserved2"] * 2 + self["IsRepeatSubPath"], 2)
        data += pack_bytes(self["reserved3"], 1)
        data += pack_bytes(self["NumberOfSubPlayItems"], 1)
        for i in self["SubPlayItems"]:
            data += i.to_bytes()
        return data


class SubPlayItem(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        self["Length"] = unpack_bytes(data, 0, 2)
        self["ClipInformationFileName"] = data[2:7].decode("utf-8")
        self["ClipCodecIdentifier"] = data[7:11].decode("utf-8")
        flags = unpack_bytes(data, 11, 4)
        self["reserved1"], flags = divmod(flags, 2 ** 5)
        self["ConnectionCondition"], self["IsMultiClipEntries"] = divmod(flags, 2)
        self["RefToSTCID"] = unpack_bytes(data, 15, 1)
        self["INTime"] = unpack_bytes(data, 16, 4)
        self["OUTTime"] = unpack_bytes(data, 20, 4)
        self["SyncPlayItemID"] = unpack_bytes(data, 24, 2)
        self["SyncStartPTS"] = unpack_bytes(data, 26, 4)

        if self["IsMultiClipEntries"]:
            self["NumberOfMultiClipEntries"] = unpack_bytes(data, 30, 1)
            self["reserved2"] = unpack_bytes(data, 31, 1)
            self["MultiClipEntries"] = []
            for i in range(self["NumberOfMultiClipEntries"]):
                self["MultiClipEntries"].append(MultiClipEntry.from_bytes(data[32 + 10 * i: 32 + 10 * (i + 1)]))

        return self

    def calculate_display_size(self):
        if self["IsMultiClipEntries"]:
            return 32 + len(self["MultiClipEntries"]) * 10 - 2
        else:
            return 30 - 2

    def update_counts(self):
        if self["IsMultiClipEntries"]:
            self["NumberOfMultiClipEntries"] = len(self["MultiClipEntries"])

    def check_constraints(self):
        super().check_constraints()
        if self["IsMultiClipEntries"]:
            assert self["NumberOfMultiClipEntries"] == len(self["MultiClipEntries"])

    def to_bytes(self):
        self.check_constraints()

        data = b""
        data += pack_bytes(self["Length"], 2)
        data += self["ClipInformationFileName"].encode("utf-8")
        data += self["ClipCodecIdentifier"].encode("utf-8")
        flags = (
                (self["reserved1"] << 5)
                + (self["ConnectionCondition"] << 1)
                + self["IsMultiClipEntries"]
        )
        data += pack_bytes(flags, 4)
        data += pack_bytes(self["RefToSTCID"], 1)
        data += pack_bytes(self["INTime"], 4)
        data += pack_bytes(self["OUTTime"], 4)
        data += pack_bytes(self["SyncPlayItemID"], 2)
        data += pack_bytes(self["SyncStartPTS"], 4)

        if self["IsMultiClipEntries"]:
            data += pack_bytes(self["NumberOfMultiClipEntries"], 1)
            data += pack_bytes(self["reserved2"], 1)
            for i in self["MultiClipEntries"]:
                data += i.to_bytes()
        return data


class MultiClipEntry(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        assert len(data) == 10
        self = cls()
        self["ClipInformationFileName"] = data[0:5].decode("utf-8")
        self["ClipCodecIdentifier"] = data[5:9].decode("utf-8")
        self["RefToSTCID"] = unpack_bytes(data, 9, 1)
        return self

    def to_bytes(self):
        self.check_constraints()

        data = b""
        data += self["ClipInformationFileName"].encode("utf-8")
        data += self["ClipCodecIdentifier"].encode("utf-8")
        data += pack_bytes(self["RefToSTCID"], 1)

        return data


class PlayListMark(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        self["Length"] = unpack_bytes(data, 0, 4)
        self["NumberOfPlayListMarks"] = unpack_bytes(data, 4, 2)
        self["PlayListMarks"] = []
        for i in range(self["NumberOfPlayListMarks"]):
            self["PlayListMarks"].append(PlayListMarkItem.from_bytes(data[6 + 14 * i: 6 + 14 * (i + 1)]))
        return self

    def calculate_display_size(self):
        return self["NumberOfPlayListMarks"] * 14 + 2

    def update_counts(self):
        self["NumberOfPlayListMarks"] = len(self["PlayListMarks"])

    def check_constraints(self):
        super().check_constraints()
        assert self["NumberOfPlayListMarks"] == len(self["PlayListMarks"])

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["Length"], 4)
        data += pack_bytes(self["NumberOfPlayListMarks"], 2)
        for i in self["PlayListMarks"]:
            data += i.to_bytes()
        return data


class PlayListMarkItem(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        assert len(data) == 14
        self = cls()
        self["reserved1"] = unpack_bytes(data, 0, 1)
        self["MarkType"] = unpack_bytes(data, 1, 1)
        self["RefToPlayItemID"] = unpack_bytes(data, 2, 2)
        self["MarkTimeStamp"] = unpack_bytes(data, 4, 4)
        self["EntryESPID"] = unpack_bytes(data, 8, 2)
        self["Duration"] = unpack_bytes(data, 10, 4)

        return self

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["reserved1"], 1)
        data += pack_bytes(self["MarkType"], 1)
        data += pack_bytes(self["RefToPlayItemID"], 2)
        data += pack_bytes(self["MarkTimeStamp"], 4)
        data += pack_bytes(self["EntryESPID"], 2)
        data += pack_bytes(self["Duration"], 4)
        return data


class MoviePlaylist:
    def __init__(self, filename=None):
        if not filename:
            self.data = MPLSHeader()
        else:
            self.load(filename)

    def load(self, filename):
        with open(filename, "rb") as f:
            data = f.read()
        self.data = MPLSHeader.from_bytes(data)

    def save(self, destination, overwrite=False):
        self.data.update_constants()
        self.data.update_addresses()
        if os.path.exists(destination) and not overwrite:
            raise FileExistsError()
        os.makedirs(os.path.dirname(destination), exist_ok=True)
        with open(destination, "wb") as f:
            f.write(self.data.to_bytes())



import os

from shinya.bd.extension_data import ExtensionData
from shinya.common.info_dict import InfoDict
from shinya.common.io import unpack_bytes, pack_bytes


class CLPIHeader(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        self["TypeIndicator"] = data[0:4].decode("utf-8")
        self["VersionNumber"] = data[4:8].decode("utf-8")
        self["SequenceInfoStartAddress"] = unpack_bytes(data, 8, 4)
        self["ProgramInfoStartAddress"] = unpack_bytes(data, 12, 4)
        self["CPIStartAddress"] = unpack_bytes(data, 16, 4)
        self["ClipMarkStartAddress"] = unpack_bytes(data, 20, 4)
        self["ExtensionDataStartAddress"] = unpack_bytes(data, 24, 4)
        self["reserved1"] = data[28:40]

        clip_info_display_size = unpack_bytes(data, 40, 4)
        sequence_info_display_size = unpack_bytes(data, self["SequenceInfoStartAddress"], 4)
        program_info_display_size = unpack_bytes(data, self["ProgramInfoStartAddress"], 4)
        cpi_display_size = unpack_bytes(data, self["CPIStartAddress"], 4)
        clip_mark_display_size = unpack_bytes(data, self["ClipMarkStartAddress"], 4)

        if self["ExtensionDataStartAddress"]:
            extension_display_size = unpack_bytes(data, self["ExtensionDataStartAddress"], 4)

        assert 40 + clip_info_display_size + 4 == self["SequenceInfoStartAddress"]
        assert self["SequenceInfoStartAddress"] + sequence_info_display_size + 4 == self["ProgramInfoStartAddress"]
        assert self["ProgramInfoStartAddress"] + program_info_display_size + 4 == self["CPIStartAddress"]
        assert self["CPIStartAddress"] + cpi_display_size + 4 == self["ClipMarkStartAddress"]

        if self["ExtensionDataStartAddress"]:
            assert self["ClipMarkStartAddress"] + clip_mark_display_size + 4 == self["ExtensionDataStartAddress"]
            assert self["ExtensionDataStartAddress"] + extension_display_size + 4 == len(data)
        else:
            assert self["ClipMarkStartAddress"] + clip_mark_display_size + 4 == len(data)

        self["ClipInfo"] = ClipInfo.from_bytes(data[40: 40 + clip_info_display_size + 4])
        self["SequenceInfo"] = SequenceInfo.from_bytes(
            data[self["SequenceInfoStartAddress"]: self["SequenceInfoStartAddress"] + sequence_info_display_size + 4])
        self["ProgramInfo"] = ProgramInfo.from_bytes(
            data[self["ProgramInfoStartAddress"]: self["ProgramInfoStartAddress"] + program_info_display_size + 4])
        self["CPI"] = CPI.from_bytes(data[self["CPIStartAddress"]: self["CPIStartAddress"] + cpi_display_size + 4])
        self["ClipMark"] = ClipMark.from_bytes(
            data[self["ClipMarkStartAddress"]:self["ClipMarkStartAddress"] + clip_mark_display_size + 4])
        if self["ExtensionDataStartAddress"]:
            self["ExtensionData"] = ExtensionData.from_bytes(
                data[self["ExtensionDataStartAddress"]: self["ExtensionDataStartAddress"] + extension_display_size + 4],
                self["ExtensionDataStartAddress"])

        assert data == self.to_bytes()
        return self

    def update_addresses(self, offset=0):
        clip_info_display_size = self["ClipInfo"].calculate_display_size()
        sequence_info_display_size = self["SequenceInfo"].calculate_display_size()
        program_info_display_size = self["ProgramInfo"].calculate_display_size()
        cpi_display_size = self["CPI"].calculate_display_size()
        clip_mark_display_size = self["ClipMark"].calculate_display_size()
        self["SequenceInfoStartAddress"] = 40 + clip_info_display_size + 4
        self["ProgramInfoStartAddress"] = self["SequenceInfoStartAddress"] + sequence_info_display_size + 4
        self["CPIStartAddress"] = self["ProgramInfoStartAddress"] + program_info_display_size + 4
        self["ClipMarkStartAddress"] = self["CPIStartAddress"] + cpi_display_size + 4
        if self["ExtensionDataStartAddress"]:
            self["ExtensionDataStartAddress"] = self["ClipMarkStartAddress"] + clip_mark_display_size + 4
        pass

    def check_constraints(self):
        clip_info_display_size = self["ClipInfo"].calculate_display_size()
        sequence_info_display_size = self["SequenceInfo"].calculate_display_size()
        program_info_display_size = self["ProgramInfo"].calculate_display_size()
        cpi_display_size = self["CPI"].calculate_display_size()
        clip_mark_display_size = self["ClipMark"].calculate_display_size()
        assert 40 + clip_info_display_size + 4 == self["SequenceInfoStartAddress"]
        assert self["SequenceInfoStartAddress"] + sequence_info_display_size + 4 == self["ProgramInfoStartAddress"]
        assert self["ProgramInfoStartAddress"] + program_info_display_size + 4 == self["CPIStartAddress"]
        assert self["CPIStartAddress"] + cpi_display_size + 4 == self["ClipMarkStartAddress"]
        if self["ExtensionDataStartAddress"]:
            assert self["ClipMarkStartAddress"] + clip_mark_display_size + 4 == self["ExtensionDataStartAddress"]

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += self["TypeIndicator"].encode("utf-8")
        data += self["VersionNumber"].encode("utf-8")
        data += pack_bytes(self["SequenceInfoStartAddress"], 4)
        data += pack_bytes(self["ProgramInfoStartAddress"], 4)
        data += pack_bytes(self["CPIStartAddress"], 4)
        data += pack_bytes(self["ClipMarkStartAddress"], 4)
        data += pack_bytes(self["ExtensionDataStartAddress"], 4)
        data += self["reserved1"]

        data += self["ClipInfo"].to_bytes()
        data += self["SequenceInfo"].to_bytes()
        data += self["ProgramInfo"].to_bytes()
        data += self["CPI"].to_bytes()
        data += self["ClipMark"].to_bytes()
        if self["ExtensionDataStartAddress"]:
            data += self["ExtensionData"].to_bytes(self["ExtensionDataStartAddress"])
        return data


class ClipInfo(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        self["Length"] = unpack_bytes(data, 0, 4)
        self["reserved1"] = unpack_bytes(data, 4, 2)
        self["ClipStreamType"] = unpack_bytes(data, 6, 1)
        self["ApplicationType"] = unpack_bytes(data, 7, 1)
        flags = unpack_bytes(data, 8, 4)
        self["reserved2"], self["IsCC5"] = divmod(flags, 2)
        self["TSRecordingRate"] = unpack_bytes(data, 12, 4)
        self["NumberOfSourcePackets"] = unpack_bytes(data, 16, 4)
        self["reserved3"] = data[20:148]
        self["TSTypeInfoBlock"] = TSTypeInfoBlock.from_bytes(data[148:180])
        if self["IsCC5"]:
            self["reserved4"] = unpack_bytes(data, 180, 1)
            self["FollowingClipStreamType"] = unpack_bytes(data, 181, 1)
            self["reserved5"] = unpack_bytes(data, 182, 4)
            self["FollowingClipInformationFileName"] = data[186:191].decode("utf-8")
            self["FollowingClipCodecIdentifier"] = data[191:195].decode("utf-8")
            self["reserved6"] = unpack_bytes(data, 195, 1)
        return self

    def calculate_display_size(self):
        if self["IsCC5"]:
            return 196 - 4
        else:
            return 180 - 4

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["Length"], 4)
        data += pack_bytes(self["reserved1"], 2)
        data += pack_bytes(self["ClipStreamType"], 1)
        data += pack_bytes(self["ApplicationType"], 1)
        data += pack_bytes(self["reserved2"] * 2 + self["IsCC5"], 4)
        data += pack_bytes(self["TSRecordingRate"], 4)
        data += pack_bytes(self["NumberOfSourcePackets"], 4)
        data += self["reserved3"]
        data += self["TSTypeInfoBlock"].to_bytes()
        if self["IsCC5"]:
            data += pack_bytes(self["reserved4"], 1)
            data += pack_bytes(self["FollowingClipStreamType"], 1)
            data += pack_bytes(self["reserved5"], 4)
            data += self["FollowingClipInformationFileName"].encode("utf-8")
            data += self["FollowingClipCodecIdentifier"].encode("utf-8")
            data += pack_bytes(self["reserved6"], 1)
        return data


class TSTypeInfoBlock(InfoDict):
    """ Specs from tsMuxer
    """

    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        self["Length"] = unpack_bytes(data, 0, 2)
        self["ValidityFlags"] = unpack_bytes(data, 2, 1)
        self["FormatIdentifier"] = data[3:7].decode("utf-8")
        self["NetworkInformation"] = data[7:16]
        self["StreamFormatName"] = data[16:32]
        return self

    def calculate_display_size(self):
        return 32 - 2

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["Length"], 2)
        data += pack_bytes(self["ValidityFlags"], 1)
        data += self["FormatIdentifier"].encode("utf-8")
        data += self["NetworkInformation"]
        data += self["StreamFormatName"]

        return data


class SequenceInfo(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):

        self = cls()
        self["Length"] = unpack_bytes(data, 0, 4)
        self["reserved1"] = unpack_bytes(data, 4, 1)
        self["NumberOfATCSequences"] = unpack_bytes(data, 5, 1)
        read_index = 6
        self["ATCSequences"] = []
        for i in range(self["NumberOfATCSequences"]):
            num_stc_seq = unpack_bytes(data, read_index + 4, 1)
            act_real_length = 6 + num_stc_seq * 14
            self["ATCSequences"].append(ATCSequence.from_bytes(data[read_index:read_index + act_real_length]))
            read_index += act_real_length

        return self

    def calculate_display_size(self):
        real_length = 6
        for i in self["ATCSequences"]:
            real_length += i.calculate_display_size()
        return real_length - 4

    def update_counts(self):
        self["NumberOfATCSequences"] = len(self["ATCSequences"])

    def check_constraints(self):
        super().check_constraints()
        assert self["NumberOfATCSequences"] == len(self["ATCSequences"])

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["Length"], 4)
        data += pack_bytes(self["reserved1"], 1)
        data += pack_bytes(self["NumberOfATCSequences"], 1)
        for i in self["ATCSequences"]:
            data += i.to_bytes()
        return data


class ATCSequence(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):

        self = cls()
        self["SPNATCStart"] = unpack_bytes(data, 0, 4)
        self["NumberOfSTCSequences"] = unpack_bytes(data, 4, 1)
        self["OffsetSTCID"] = unpack_bytes(data, 5, 1)
        self["STCSequences"] = []
        for i in range(self["NumberOfSTCSequences"]):
            self["STCSequences"].append(STCSequence.from_bytes(data[6 + 14 * i:6 + 14 * (i + 1)]))

        return self

    def calculate_display_size(self):
        return 6 + 14 * len(self["STCSequences"])

    def update_counts(self):
        self["NumberOfSTCSequences"] = len(self["STCSequences"])

    def check_constraints(self):
        super().check_constraints()
        assert self["NumberOfSTCSequences"] == len(self["STCSequences"])

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["SPNATCStart"], 4)
        data += pack_bytes(self["NumberOfSTCSequences"], 1)
        data += pack_bytes(self["OffsetSTCID"], 1)
        for i in self["STCSequences"]:
            data += i.to_bytes()
        return data


class STCSequence(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        self["PCRPID"] = unpack_bytes(data, 0, 2)
        self["SPNSTCStart"] = unpack_bytes(data, 2, 4)
        self["PresentationStartTime"] = unpack_bytes(data, 6, 4)
        self["PresentationEndTime"] = unpack_bytes(data, 10, 4)

        return self

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["PCRPID"], 2)
        data += pack_bytes(self["SPNSTCStart"], 4)
        data += pack_bytes(self["PresentationStartTime"], 4)
        data += pack_bytes(self["PresentationEndTime"], 4)
        return data


class ProgramInfo(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):

        self = cls()
        self["Length"] = unpack_bytes(data, 0, 4)
        self["reserved1"] = unpack_bytes(data, 4, 1)
        self["NumberOfPrograms"] = unpack_bytes(data, 5, 1)
        read_index = 6
        self["Programs"] = []
        for i in range(self["NumberOfPrograms"]):
            num_stream_ps = unpack_bytes(data, read_index + 6, 1)
            program_offset = 8
            for j in range(num_stream_ps):
                program_offset += 2
                program_offset += unpack_bytes(data, read_index + program_offset, 1) + 1
            self["Programs"].append(Program.from_bytes(data[read_index:read_index + program_offset]))
            read_index += program_offset

        return self

    def calculate_display_size(self):
        real_length = 6
        for i in self["Programs"]:
            real_length += i.calculate_display_size()
        return real_length - 4

    def update_counts(self):
        self["NumberOfPrograms"] = len(self["Programs"])

    def check_constraints(self):
        super().check_constraints()
        assert self["NumberOfPrograms"] == len(self["Programs"])

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["Length"], 4)
        data += pack_bytes(self["reserved1"], 1)
        data += pack_bytes(self["NumberOfPrograms"], 1)
        for i in self["Programs"]:
            data += i.to_bytes()
        return data


class Program(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):

        self = cls()
        self["SPNProgramSequenceStart"] = unpack_bytes(data, 0, 4)
        self["ProgramMapPID"] = unpack_bytes(data, 4, 2)
        self["NumberOfStreamsInPS"] = unpack_bytes(data, 6, 1)
        self["NumberOfGroups"] = unpack_bytes(data, 7, 1)
        read_index = 8
        self["StreamsInPS"] = []
        for i in range(self["NumberOfStreamsInPS"]):
            streams_in_ps = InfoDict()
            streams_in_ps["StreamPID"] = unpack_bytes(data, read_index, 2)
            read_index += 2
            scinfo_display_size = unpack_bytes(data, read_index, 1)
            streams_in_ps["StreamCodingInfo"] = StreamCodingInfo.from_bytes(
                data[read_index:read_index + scinfo_display_size + 1])
            self["StreamsInPS"].append(streams_in_ps)
            read_index += scinfo_display_size + 1

        return self

    def calculate_display_size(self):
        real_length = 8
        for i in self["StreamsInPS"]:
            real_length += 2
            real_length += i["StreamCodingInfo"].calculate_display_size() + 1
        return real_length

    def update_counts(self):
        self["NumberOfStreamsInPS"] = len(self["StreamsInPS"])

    def check_constraints(self):
        super().check_constraints()
        assert self["NumberOfStreamsInPS"] == len(self["StreamsInPS"])

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["SPNProgramSequenceStart"], 4)
        data += pack_bytes(self["ProgramMapPID"], 2)
        data += pack_bytes(self["NumberOfStreamsInPS"], 1)
        data += pack_bytes(self["NumberOfGroups"], 1)
        for i in self["StreamsInPS"]:
            data += pack_bytes(i["StreamPID"], 2)
            data += i["StreamCodingInfo"].to_bytes()
        return data


class StreamCodingInfo(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):

        self = cls()
        self["Length"] = unpack_bytes(data, 0, 1)
        self["StreamCodingType"] = unpack_bytes(data, 1, 1)
        if self["StreamCodingType"] in [0x01, 0x02, 0x1B, 0xEA]:
            self["VideoFormat"], self["FrameRate"] = divmod(unpack_bytes(data, 2, 1), 2 ** 4)
            flags = unpack_bytes(data, 3, 1)
            self["VideoAspect"], flags = divmod(flags, 2 ** 4)
            self["reserved1"], flags = divmod(flags, 2 ** 2)
            self["OCFlag"], self["reserved2"] = divmod(flags, 2)
            self["reserved3"] = unpack_bytes(data, 4, 2)
            self["padding"] = data[6:]
        elif self["StreamCodingType"] in [0x24]:
            self["VideoFormat"], self["FrameRate"] = divmod(unpack_bytes(data, 2, 1), 2 ** 4)
            flags = unpack_bytes(data, 3, 1)
            self["VideoAspect"], flags = divmod(flags, 2 ** 4)
            self["reserved4"], flags = divmod(flags, 2 ** 2)
            self["OCFlag"], self["CRFlag"] = divmod(flags, 2)
            self["DynamicRangeType"], self["ColorSpace"] = divmod(unpack_bytes(data, 4, 1), 2 ** 4)
            self["HDRPlusFlag"], self["reserved5"] = divmod(unpack_bytes(data, 5, 1), 2 ** 7)
            self["padding"] = data[6:]
        elif self["StreamCodingType"] in [0x03, 0x04, 0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0xA1, 0xA2]:
            self["AudioFormat"], self["SampleRate"] = divmod(unpack_bytes(data, 2, 1), 2 ** 4)
            self["Language"] = data[3:6].decode("utf-8")
            self["padding"] = data[6:]
        elif self["StreamCodingType"] in [0x90, 0x91]:
            self["Language"] = data[2:5].decode("utf-8")
            self["padding"] = data[5:]
        elif self["StreamCodingType"] in [0x92]:
            self["CharCode"] = unpack_bytes(data, 2, 1)
            self["Language"] = data[3:6].decode("utf-8")
            self["padding"] = data[6:]

        return self

    def calculate_display_size(self):
        return self["Length"]

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["Length"], 1)
        data += pack_bytes(self["StreamCodingType"], 1)

        if self["StreamCodingType"] in [0x01, 0x02, 0x1B, 0xEA]:
            data += pack_bytes((self["VideoFormat"] << 4) + self["FrameRate"], 1)
            data += pack_bytes(
                (self["VideoAspect"] << 4) + (self["reserved1"] << 2) + (self["OCFlag"] << 1) + self["reserved2"], 1)
            data += pack_bytes(self["reserved3"], 2)
            data += self["padding"]
        elif self["StreamCodingType"] in [0x24]:
            data += pack_bytes((self["VideoFormat"] << 4) + self["FrameRate"], 1)
            data += pack_bytes(
                (self["VideoAspect"] << 4) + (self["reserved4"] << 2) + (self["OCFlag"] << 1) + self["CRFlag"], 1)
            data += pack_bytes((self["DynamicRangeType"] << 4) + self["ColorSpace"], 1)
            data += pack_bytes((self["HDRPlusFlag"] << 7) + self["reserved5"], 1)
            data += self["padding"]
        elif self["StreamCodingType"] in [0x03, 0x04, 0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0xA1, 0xA2]:
            data += pack_bytes((self["AudioFormat"] << 4) + self["SampleRate"], 1)
            data += self["Language"].encode("utf-8")
            data += self["padding"]
        elif self["StreamCodingType"] in [0x90, 0x91]:
            data += self["Language"].encode("utf-8")
            data += self["padding"]
        elif self["StreamCodingType"] in [0x92]:
            data += pack_bytes(self["CharCode"], 1)
            data += self["Language"].encode("utf-8")
            data += self["padding"]

        assert len(data) == self["Length"] + 1
        return data


class CPI(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        self["Length"] = unpack_bytes(data, 0, 4)
        if self["Length"]:
            flags = unpack_bytes(data, 4, 2)
            self["reserved1"], self["CPIType"] = divmod(flags, 2 ** 4)
            # EPMap starts here, at relative real position 6, display position 2
            self["reserved2"] = unpack_bytes(data, 6, 1)
            self["NumberOfStreamPIDEntries"] = unpack_bytes(data, 7, 1)
            # Do not create a new class for StreamPIDEntries, since its data is not continuous, conversation should
            # be handled here instead of calling subclass methods
            self["StreamPIDEntries"] = []
            # meta-info block contains 12 Bytes
            for i in range(self["NumberOfStreamPIDEntries"]):
                spid_entry = InfoDict()
                # read_offset is static
                read_offset = 8 + 12 * i
                tmp = unpack_bytes(data, read_offset, 8)
                spid_entry["StreamPID"], tmp = divmod(tmp, 2 ** 48)
                spid_entry["reserved3"], tmp = divmod(tmp, 2 ** 38)
                spid_entry["EPStreamType"], tmp = divmod(tmp, 2 ** 34)
                spid_entry["NumberOfEPCoarseEntries"], spid_entry["NumberOfEPFineEntries"] = divmod(tmp, 2 ** 18)
                spid_entry["EPMapForOneStreamPIDStartAddress"] = unpack_bytes(data, read_offset + 8, 4)
                self["StreamPIDEntries"].append(spid_entry)

            # block_start_address is dynamic
            current_address = 8 + 12 * self["NumberOfStreamPIDEntries"]
            for spid_entry in self["StreamPIDEntries"]:
                # Each block is of size 4 + 8 * NumberOfEPCoarseEntries + 4 * NumberOfEPFineEntries
                # address formula 1
                assert spid_entry["EPMapForOneStreamPIDStartAddress"] == current_address - 6
                spid_entry["EPFineTableStartAddress"] = unpack_bytes(data, current_address, 4)
                current_address += 4
                spid_entry["EPCoarseEntries"] = []
                for j in range(spid_entry["NumberOfEPCoarseEntries"]):
                    spid_entry["EPCoarseEntries"].append(EPCoarseEntry.from_bytes(
                        data[current_address:current_address + 8]
                    ))
                    current_address += 8

                # address formula 2
                assert spid_entry["EPFineTableStartAddress"] + \
                       spid_entry["EPMapForOneStreamPIDStartAddress"] == current_address - 6
                spid_entry["EPFineEntries"] = []
                for j in range(spid_entry["NumberOfEPFineEntries"]):
                    spid_entry["EPFineEntries"].append(EPFineEntry.from_bytes(
                        data[current_address:current_address + 4]
                    ))
                    current_address += 4

        return self

    def calculate_display_size(self):
        if self["Length"]:
            real_length = 8 + 12 * len(self["StreamPIDEntries"])
            for spid_entry in self["StreamPIDEntries"]:
                real_length += 4 + 8 * len(spid_entry["EPCoarseEntries"]) + 4 * len(spid_entry["EPFineEntries"])
            return real_length - 4
        else:
            return 0

    def update_counts(self):
        self["NumberOfStreamPIDEntries"] = len(self["StreamPIDEntries"])
        for spid_entry in self["StreamPIDEntries"]:
            spid_entry["NumberOfEPCoarseEntries"] = len(spid_entry["EPCoarseEntries"])
            spid_entry["NumberOfEPFineEntries"] = len(spid_entry["EPFineEntries"])
        current_address = 8 + 12 * len(self["StreamPIDEntries"])
        for spid_entry in self["StreamPIDEntries"]:
            spid_entry["EPMapForOneStreamPIDStartAddress"] = current_address - 6
            current_address += 4 + 8 * len(spid_entry["EPCoarseEntries"])
            spid_entry["EPFineTableStartAddress"] = current_address - 6 - spid_entry["EPMapForOneStreamPIDStartAddress"]
            current_address += 4 * len(spid_entry["EPFineEntries"])

    def check_constraints(self):
        super().check_constraints()
        assert self["NumberOfStreamPIDEntries"] == len(self["StreamPIDEntries"])
        for spid_entry in self["StreamPIDEntries"]:
            assert spid_entry["NumberOfEPCoarseEntries"] == len(spid_entry["EPCoarseEntries"])
            assert spid_entry["NumberOfEPFineEntries"] == len(spid_entry["EPFineEntries"])
        current_address = 8 + 12 * len(self["StreamPIDEntries"])
        for spid_entry in self["StreamPIDEntries"]:
            assert spid_entry["EPMapForOneStreamPIDStartAddress"] == current_address - 6
            current_address += 4 + 8 * len(spid_entry["EPCoarseEntries"])
            assert spid_entry["EPFineTableStartAddress"] == current_address - 6 - spid_entry[
                "EPMapForOneStreamPIDStartAddress"]
            current_address += 4 * len(spid_entry["EPFineEntries"])

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["Length"], 4)
        if self["Length"]:
            data += pack_bytes((self["reserved1"] << 4) + self["CPIType"], 2)
            data += pack_bytes(self["reserved2"], 1)
            data += pack_bytes(self["NumberOfStreamPIDEntries"], 1)
            for spid_entry in self["StreamPIDEntries"]:
                data += pack_bytes((spid_entry["StreamPID"] << 48)
                                   + (spid_entry["reserved3"] << 38)
                                   + (spid_entry["EPStreamType"] << 34)
                                   + (spid_entry["NumberOfEPCoarseEntries"] << 18)
                                   + spid_entry["NumberOfEPFineEntries"], 8)
                data += pack_bytes(spid_entry["EPMapForOneStreamPIDStartAddress"], 4)
            for spid_entry in self["StreamPIDEntries"]:
                data += pack_bytes(spid_entry["EPFineTableStartAddress"], 4)
                for i in spid_entry["EPCoarseEntries"]:
                    data += i.to_bytes()
                for i in spid_entry["EPFineEntries"]:
                    data += i.to_bytes()
        return data


class EPCoarseEntry(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        flags = unpack_bytes(data, 0, 4)
        self["RefToEPFineID"], self["PTSEPCoarse"] = divmod(flags, 2 ** 14)
        self["SPNEPCoarse"] = unpack_bytes(data, 4, 4)
        return self

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes((self["RefToEPFineID"] << 14) + self["PTSEPCoarse"], 4)
        data += pack_bytes(self["SPNEPCoarse"], 4)
        return data


class EPFineEntry(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        flags = unpack_bytes(data, 0, 4)
        self["IsAngleChangePoint"], flags = divmod(flags, 2 ** 31)
        self["IEndPositionOffset"], flags = divmod(flags, 2 ** 28)
        self["PTSEPFine"], self["SPNEPFine"] = divmod(flags, 2 ** 17)
        return self

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(
            (self["IsAngleChangePoint"] << 31) + (self["IEndPositionOffset"] << 28) + (self["PTSEPFine"] << 17) +
            self["SPNEPFine"], 4)
        return data


class ClipMark(InfoDict):
    """ No specs available
    """

    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        self["Length"] = unpack_bytes(data, 0, 4)
        if self["Length"] != 0:
            self["Data"] = data[4:]
        return self

    def calculate_display_size(self):
        if self["Length"] != 0:
            return len(self["data"])
        else:
            return 0

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["Length"], 4)
        if self["Length"] != 0:
            data += self["Data"]

        return data


class ClipInformationFile:
    def __init__(self, filename=None):
        if not filename:
            self.data = CLPIHeader()
        else:
            self.load(filename)

    def load(self, filename):
        with open(filename, "rb") as f:
            data = f.read()
        self.data = CLPIHeader.from_bytes(data)

    def save(self, destination, overwrite=False):
        self.data.update_constants()
        self.data.update_addresses()
        if os.path.exists(destination) and not overwrite:
            raise FileExistsError()
        os.makedirs(os.path.dirname(destination), exist_ok=True)
        with open(destination, "wb") as f:
            f.write(self.data.to_bytes())



import os
from copy import deepcopy

from lxml import etree

from shinya.bd.mpls import MoviePlaylist


class MatroskaXMLChapter:
    """
    Minimal implementation of xml chapter generated from blu-ray discs
    """

    def __init__(self, chapters):
        assert isinstance(chapters, Chapter)
        self.chapters = chapters
        self.xml_chapter = etree.Element('Chapters')
        edition_entry = etree.SubElement(self.xml_chapter, 'EditionEntry')
        edition_flag_hidden = etree.SubElement(edition_entry, 'EditionFlagHidden')
        edition_flag_hidden.text = "0"
        edition_flag_default = etree.SubElement(edition_entry, 'EditionFlagDefault')
        edition_flag_default.text = "1"
        for index, chapter_entry in enumerate(self.chapters):
            chapter_atom = etree.SubElement(edition_entry, "ChapterAtom")
            chapter_time_start = etree.SubElement(chapter_atom, "ChapterTimeStart")
            chapter_time_start.text = chapter_entry.time_str
            chapter_flag_hidden = etree.SubElement(chapter_atom, "ChapterFlagHidden")
            chapter_flag_hidden.text = "0"
            chapter_flag_enabled = etree.SubElement(chapter_atom, "ChapterFlagEnabled")
            chapter_flag_enabled.text = "1"
            chapter_display = etree.SubElement(chapter_atom, "ChapterDisplay")
            chapter_string = etree.SubElement(chapter_display, "ChapterString")
            if chapter_entry.name:
                chapter_string.text = chapter_entry.name
            else:
                chapter_string.text = f"Chapter {index + 1:02d}"
            chapter_language = etree.SubElement(chapter_display, "ChapterLanguage")
            chapter_language.text = chapter_entry.language

    def export(self, destination, overwrite=False):
        if os.path.exists(destination) and not overwrite:
            raise FileExistsError()
        os.makedirs(os.path.dirname(destination), exist_ok=True)
        with open(destination, "wb") as f:
            f.write(etree.tostring(self.xml_chapter, encoding="utf-8",
                                   doctype='<!DOCTYPE Chapters SYSTEM "matroskachapters.dtd">',
                                   xml_declaration=True,
                                   pretty_print=True))


class ChapterEntry:
    def __init__(self, time_sec, language="eng", name=None):
        assert time_sec >= 0.
        assert len(language) == 3
        self._time_sec = time_sec
        self.language = language
        self.name = name

    @property
    def time_sec(self):
        return self._time_sec

    @time_sec.setter
    def time_sec(self, value):
        if value < 0.:
            raise ValueError("Time stamp of a chapter cannot be negative.")
        self._time_sec = value

    @property
    def time_str(self):
        seconds = self.time_sec
        hour = int(seconds) // 3600
        seconds %= 3600
        minute = int(seconds) // 60
        seconds %= 60
        seconds_int = int(seconds)
        seconds_decimals = int((seconds % 1) * 10 ** 6)
        return f"{hour:02d}:{minute:02d}:{seconds_int:02d}.{seconds_decimals:06d}"

    def __str__(self):
        return f"{self.time_str}\t{self.language}" + (f"{self.name}" if self.name else "")

    def __repr__(self):
        return f"ChapterEntry({self.time_str}, {self.language}, {self.name})"


class Chapter:
    def __init__(self, data=None, end_time=None):
        if not data:
            self.data = []
        else:
            self.data = data
        self._end_time = end_time
        self.check_data()

    @property
    def end_time(self):
        return self._end_time

    @end_time.setter
    def end_time(self, value):
        for c in self.data:
            if value < c.time_sec:
                raise ValueError("End time of the stream must be larger than all chapters.")
        self._end_time = value

    @classmethod
    def from_mpls(cls, filename):
        """

        Args:
            filename: filename of the mpls file

        Returns:
            A list of tuples, [(play_item_ID: int, chapters: Chapter)..]
        """
        mpls = MoviePlaylist(filename)
        result = []
        play_items = mpls.data["PlayList"]["PlayItems"]
        playlist_marks = mpls.data["PlayListMark"]["PlayListMarks"]
        playlist_marks_dict = {}
        for playlist_mark in playlist_marks:
            # entry-mark
            if playlist_mark["MarkType"] == 1:
                ref_to_pi = playlist_mark["RefToPlayItemID"]
                time_raw = playlist_mark["MarkTimeStamp"]
                if ref_to_pi in playlist_marks_dict:
                    playlist_marks_dict[ref_to_pi].append(time_raw)
                else:
                    playlist_marks_dict[ref_to_pi] = [time_raw]
        for play_item_index, play_item in enumerate(play_items):
            if play_item_index not in playlist_marks_dict:
                continue
            in_time = play_item["INTime"]
            end_time = (play_item["OUTTime"] - in_time) / 45000
            chapter_data = []
            first_chapter_time = playlist_marks_dict[play_item_index][0] - in_time
            if first_chapter_time > 0:
                chapter_data.append(ChapterEntry(0.))
            if first_chapter_time < 0:
                raise ValueError("First chapter time is earlier than play item in-time, this is considered an error.")

            for raw_time in playlist_marks_dict[play_item_index]:
                chapter_data.append(ChapterEntry((raw_time - in_time) / 45000))

            result.append((play_item["ClipInformationFileName"], Chapter(chapter_data, end_time)))

        return result

    def check_data(self):
        assert isinstance(self.data, list)
        # the first chapter must always be at the beginning
        if self.data:
            assert self.data[0].time_sec == 0.
        for c in self.data:
            assert isinstance(c, ChapterEntry)

    def export(self, destination, export_format="xml"):
        if export_format == "xml":
            handler = MatroskaXMLChapter(self)
            handler.export(destination)
        else:
            raise NotImplementedError(f"Exporting chapters to format {export_format} is not supported yet.")

    def __getitem__(self, item):
        if isinstance(item, slice):
            indices = range(*item.indices(len(self.data)))
            # preserves all other attributes
            result = deepcopy(self)
            if len(indices) == 0:
                result.data = []
                return result
            new_data = []
            init_offset = self.data[indices[0]].time_sec
            for i in indices:
                new_chapter_entry = deepcopy(self.data[i])
                new_chapter_entry.time_sec -= init_offset
                new_data.append(new_chapter_entry)
            result.data = new_data
            result.end_time -= init_offset
            return result
        return self.data[item]

    def __add__(self, other):
        if isinstance(other, Chapter):
            # attributes will follow the current object, except for data and end_time
            result = deepcopy(self)
            new_offset = self.end_time
            for c in other.data:
                new_chapter_entry = deepcopy(c)
                new_chapter_entry.time_sec += new_offset
                result.data.append(new_chapter_entry)
            result.end_time += other.end_time
            return result
        elif isinstance(other, int) or isinstance(other, float):
            if other == 0.:
                return self
            else:
                raise NotImplementedError()
        else:
            raise ValueError()

    def __radd__(self, other):
        return self.__add__(other)

    def __repr__(self):
        return f"Chapter({self.data}, {self.end_time})"



import struct

FORMAT_CHAR = {1: ">B", 2: ">H", 4: ">I", 8: ">Q"}


def unpack_bytes(data, offset, length):
    result, = struct.unpack(FORMAT_CHAR[length], data[offset:offset + length])
    return result


def pack_bytes(data, length):
    return struct.pack(FORMAT_CHAR[length], data)




from abc import abstractmethod
from collections import OrderedDict


class InfoDict(OrderedDict):
    def __init__(self):
        super().__init__()

    @classmethod
    @abstractmethod
    def from_bytes(cls, data, **kwargs):
        raise NotImplementedError()

    def check_constraints(self):
        if "Length" in self:
            assert self["Length"] == self.calculate_display_size()

    def calculate_display_size(self):
        raise NotImplementedError()

    def update_counts(self):
        pass

    def update_addresses(self, offset=0):
        pass

    def update_constants(self):
        for key in self.keys():
            if isinstance(self[key], InfoDict):
                self[key].update_constants()
            elif isinstance(self[key], list):
                for item in self[key]:
                    if isinstance(item, InfoDict):
                        item.update_constants()
        self.update_counts()
        if "Length" in self:
            self["Length"] = self.calculate_display_size()

    @abstractmethod
    def to_bytes(self, **kwargs):
        raise NotImplementedError()



import os

from shinya.bd.extension_data import ExtensionData
from shinya.common.info_dict import InfoDict
from shinya.common.io import unpack_bytes, pack_bytes


class INDXHeader(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        self["TypeIndicator"] = data[0:4].decode("utf-8")
        self["VersionNumber"] = data[4:8].decode("utf-8")
        self["IndexesStartAddress"] = unpack_bytes(data, 8, 4)
        self["ExtensionDataStartAddress"] = unpack_bytes(data, 12, 4)
        self["reserved1"] = data[16:40]

        appinfo_display_size = unpack_bytes(data, 40, 4)
        indexes_display_size = unpack_bytes(data, self["IndexesStartAddress"], 4)

        if self["ExtensionDataStartAddress"]:
            extension_display_size = unpack_bytes(data, self["ExtensionDataStartAddress"], 4)

        assert appinfo_display_size == 34
        assert self["IndexesStartAddress"] == 78
        if self["ExtensionDataStartAddress"]:
            assert (self["IndexesStartAddress"] + indexes_display_size + 4 == self["ExtensionDataStartAddress"])
            assert self["ExtensionDataStartAddress"] + extension_display_size + 4 == len(data)
        else:
            assert self["IndexesStartAddress"] + indexes_display_size + 4 == len(data)

        self["AppInfoBDMV"] = AppInfoBDMV.from_bytes(data[40: 40 + appinfo_display_size + 4])
        self["Indexes"] = Indexes.from_bytes(
            data[self["IndexesStartAddress"]: self["IndexesStartAddress"] + indexes_display_size + 4])
        if self["ExtensionDataStartAddress"]:
            self["ExtensionData"] = ExtensionData.from_bytes(
                data[self["ExtensionDataStartAddress"]: self["ExtensionDataStartAddress"] + extension_display_size + 4],
                self["ExtensionDataStartAddress"])

        assert data == self.to_bytes()
        return self

    def update_addresses(self, offset=0):
        if self["ExtensionDataStartAddress"]:
            self["ExtensionDataStartAddress"] = 40 + self['MovieObjects']['Length'] + 4

    def check_constraints(self):
        assert self["AppInfoBDMV"].calculate_display_size() == 34
        assert self["IndexesStartAddress"] == 78
        if self["ExtensionDataStartAddress"]:
            indexes_display_size = self["Indexes"].calculate_display_size()
            assert (self["IndexesStartAddress"] + indexes_display_size + 4 == self["ExtensionDataStartAddress"])

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += self["TypeIndicator"].encode("utf-8")
        data += self["VersionNumber"].encode("utf-8")
        data += pack_bytes(self["IndexesStartAddress"], 4)
        data += pack_bytes(self["ExtensionDataStartAddress"], 4)
        data += self["reserved1"]

        data += self["AppInfoBDMV"].to_bytes()
        data += self["Indexes"].to_bytes()
        if self["ExtensionDataStartAddress"]:
            data += self["ExtensionData"].to_bytes(self["ExtensionDataStartAddress"])
        return data


class AppInfoBDMV(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        assert len(data) == 38

        self = cls()
        self["Length"] = unpack_bytes(data, 0, 4)
        flags = unpack_bytes(data, 4, 1)
        self["reserved1"] = flags >> 7 & 1
        self["InitialOutputModePreference"] = flags >> 6 & 1
        self["SSContentExistFlag"] = flags >> 5 & 1
        self["reserved2"] = flags >> 4 & 1
        self["InitialDynamicRangeType"] = flags % 2 ** 4
        flags = unpack_bytes(data, 5, 1)
        self["VideoFormat"], self["FrameRate"] = divmod(flags, 2 ** 4)
        self["UserData"] = data[6:]

        return self

    def calculate_display_size(self):
        return 34

    def to_bytes(self):
        self.check_constraints()

        data = b""
        data += pack_bytes(self["Length"], 4)
        flags = (
                (self["reserved1"] << 7)
                + (self["InitialOutputModePreference"] << 6)
                + (self["SSContentExistFlag"] << 5)
                + (self["reserved2"] << 4)
                + self["InitialDynamicRangeType"]
        )
        data += pack_bytes(flags, 1)
        data += pack_bytes((self["VideoFormat"] << 4) + self["FrameRate"], 1)
        data += self["UserData"]

        return data


class Indexes(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        self["Length"] = unpack_bytes(data, 0, 4)
        self["FirstPlaybackTitle"] = Title.from_bytes(data[4:16])
        self["TopMenuTitle"] = Title.from_bytes(data[16:28])
        self["NumberOfTitles"] = unpack_bytes(data, 28, 2)
        self["Titles"] = []
        for i in range(self["NumberOfTitles"]):
            self["Titles"].append(Title.from_bytes(data[30 + i * 12:30 + (i + 1) * 12]))

        return self

    def calculate_display_size(self):
        return 30 + 12 * len(self["Titles"]) - 4

    def update_counts(self):
        self["NumberOfTitles"] = len(self["Titles"])

    def check_constraints(self):
        super().check_constraints()
        assert self["NumberOfTitles"] == len(self["Titles"])

    def to_bytes(self):
        self.check_constraints()

        data = b""
        data += pack_bytes(self["Length"], 4)
        data += self["FirstPlaybackTitle"].to_bytes()
        data += self["TopMenuTitle"].to_bytes()
        data += pack_bytes(self["NumberOfTitles"], 2)
        for i in self["Titles"]:
            data += i.to_bytes()

        return data


class Title(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        assert len(data) == 12

        self = cls()
        flags = unpack_bytes(data, 0, 4)
        self["ObjectType"], flags = divmod(flags, 2 ** 30)
        self["AccessType"], self["reserved1"] = divmod(flags, 2 ** 28)
        flags = unpack_bytes(data, 4, 2)
        self["PlaybackType"], self["reserved2"] = divmod(flags, 2 ** 14)

        if self["ObjectType"] == 1:
            self["RefToMovieObjectID"] = unpack_bytes(data, 6, 2)
            self["reserved3"] = unpack_bytes(data, 8, 4)
        else:
            self["RefToBDJObjectID"] = data[6:11].decode("utf-8")
            self["reserved4"] = unpack_bytes(data, 11, 1)

        return self

    def to_bytes(self):
        self.check_constraints()

        data = b""
        data += pack_bytes((self["ObjectType"] << 30) + (self["AccessType"] << 28) + self["reserved1"], 4)
        data += pack_bytes((self["PlaybackType"] << 14) + self["reserved2"], 2)

        if self["ObjectType"] == 1:
            data += pack_bytes(self["RefToMovieObjectID"], 2)
            data += pack_bytes(self["reserved3"], 4)
        else:
            data += self["RefToBDJObjectID"].encode("utf-8")
            data += pack_bytes(self["reserved4"], 1)

        return data


class IndexTableFile:
    def __init__(self, filename=None):
        if not filename:
            self.data = INDXHeader()
        else:
            self.load(filename)

    def load(self, filename):
        with open(filename, "rb") as f:
            data = f.read()
        self.data = INDXHeader.from_bytes(data)

    def save(self, destination, overwrite=False):
        self.data.update_constants()
        self.data.update_addresses()
        if os.path.exists(destination) and not overwrite:
            raise FileExistsError()
        os.makedirs(os.path.dirname(destination), exist_ok=True)
        with open(destination, "wb") as f:
            f.write(self.data.to_bytes())



import os
from enum import IntEnum

from shinya.bd.extension_data import ExtensionData
from shinya.common.info_dict import InfoDict
from shinya.common.io import unpack_bytes, pack_bytes


class MOBJHeader(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        self["TypeIndicator"] = data[0:4].decode("utf-8")
        self["VersionNumber"] = data[4:8].decode("utf-8")
        self["ExtensionDataStartAddress"] = unpack_bytes(data, 8, 4)
        self["reserved1"] = data[12:40]

        movie_object_length = unpack_bytes(data, 40, 4)
        self['MovieObjects'] = MovieObjects.from_bytes(data[40: 40 + movie_object_length + 4])

        if self["ExtensionDataStartAddress"]:
            assert (self["PlayListMarkStartAddress"] + movie_object_length + 4 == self[
                "ExtensionDataStartAddress"])
            extension_display_size = unpack_bytes(data, self["ExtensionDataStartAddress"], 4)
            assert self["ExtensionDataStartAddress"] + extension_display_size + 4 == len(data)
            self["ExtensionData"] = ExtensionData.from_bytes(
                data[self["ExtensionDataStartAddress"]: self["ExtensionDataStartAddress"] + extension_display_size + 4],
                self["ExtensionDataStartAddress"])
        else:
            assert 40 + movie_object_length + 4 == len(data)

        assert data == self.to_bytes()
        return self

    def update_addresses(self, offset=0):
        if self["ExtensionDataStartAddress"]:
            self["ExtensionDataStartAddress"] = 40 + self['MovieObjects']['Length'] + 4

    def check_constraints(self):
        if self["ExtensionDataStartAddress"]:
            movie_object_length = self["MovieObjects"].calculate_display_size()
            assert (self["PlayListMarkStartAddress"] + movie_object_length + 4 == self["ExtensionDataStartAddress"])

    def to_bytes(self):
        self.check_constraints()
        data = b""
        data += self["TypeIndicator"].encode("utf-8")
        data += self["VersionNumber"].encode("utf-8")
        data += pack_bytes(self["ExtensionDataStartAddress"], 4)
        data += self["reserved1"]

        data += self["MovieObjects"].to_bytes()
        if self["ExtensionDataStartAddress"]:
            data += self["ExtensionData"].to_bytes(self["ExtensionDataStartAddress"])
        return data


class MovieObjects(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        self["Length"] = unpack_bytes(data, 0, 4)
        self["reserved1"] = unpack_bytes(data, 4, 4)
        self["NumberOfMobjs"] = unpack_bytes(data, 8, 2)

        self["Mobjs"] = []
        read_index = 10
        for i in range(self["NumberOfMobjs"]):
            n_navi_cmds = unpack_bytes(data, read_index + 2, 2)
            mobj_length = 4 + n_navi_cmds * 12
            self["Mobjs"].append(Mobj.from_bytes(data[read_index:read_index + mobj_length]))
            read_index += mobj_length
        return self

    def calculate_display_size(self):
        real_length = 10
        for i in self["Mobjs"]:
            real_length += i.calculate_display_size()
        return real_length - 4

    def update_counts(self):
        self["NumberOfMobjs"] = len(self["Mobjs"])

    def check_constraints(self):
        super().check_constraints()
        assert self["NumberOfMobjs"] == len(self["Mobjs"])

    def to_bytes(self, **kwargs):
        self.check_constraints()
        data = b""
        data += pack_bytes(self["Length"], 4)
        data += pack_bytes(self["reserved1"], 4)
        data += pack_bytes(self["NumberOfMobjs"], 2)
        for i in self["Mobjs"]:
            data += i.to_bytes()

        return data


class Mobj(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        self = cls()
        flags = unpack_bytes(data, 0, 2)
        self["ResumeIntentionFlag"] = flags >> 15 & 1
        self["MenuCallMask"] = flags >> 14 & 1
        self["TitleSearchMask"] = flags >> 13 & 1
        self["reserved1"] = flags % 2 ** 13
        self["NumberOfNavigationCommands"] = unpack_bytes(data, 2, 2)

        self["NavigationCommands"] = []
        for i in range(self["NumberOfNavigationCommands"]):
            self["NavigationCommands"].append(NavigationCommand.from_bytes(data[4 + i * 12:4 + (i + 1) * 12]))
        return self

    def calculate_display_size(self):
        return 4 + 12 * len(self["NavigationCommands"])

    def update_counts(self):
        self["NumberOfNavigationCommands"] = len(self["NavigationCommands"])

    def check_constraints(self):
        super().check_constraints()
        assert self["NumberOfNavigationCommands"] == len(self["NavigationCommands"])

    def to_bytes(self, **kwargs):
        self.check_constraints()
        data = b""
        data += pack_bytes(
            self["ResumeIntentionFlag"] * 2 ** 15 + self["MenuCallMask"] * 2 ** 14 + self["TitleSearchMask"] * 2 ** 13 +
            self["reserved1"], 2)
        data += pack_bytes(self["NumberOfNavigationCommands"], 2)
        for i in self["NavigationCommands"]:
            data += i.to_bytes()

        return data


class CommandGroup(IntEnum):
    BRANCH = 0
    CMP = 1
    SET = 2


class BranchCommandSubGroup(IntEnum):
    GOTO = 0
    JUMP = 1
    PLAY = 2


class SetCommandSubGroup(IntEnum):
    SET = 0
    SETSYSTEM = 1


class NavigationCommand(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, **kwargs):
        assert len(data) == 12
        self = cls()
        flags = unpack_bytes(data, 0, 1)
        self["OperandCount"], flags = divmod(flags, 2 ** 5)
        self["CommandGroup"], self["CommandSubGroup"] = divmod(flags, 2 ** 3)
        flags = unpack_bytes(data, 1, 1)
        self["DestinationImmediateValueFlag"] = flags >> 7 & 1
        self["SourceImmediateValueFlag"] = flags >> 6 & 1
        self["reserved1"], self["BranchOption"] = divmod(flags % 2 ** 6, 2 ** 4)
        flags = unpack_bytes(data, 2, 1)
        self["reserved2"], self["CompareOption"] = divmod(flags, 2 ** 4)
        flags = unpack_bytes(data, 3, 1)
        self["reserved3"], self["SetOption"] = divmod(flags, 2 ** 5)
        self["Destination"] = unpack_bytes(data, 4, 4)
        self["Source"] = unpack_bytes(data, 8, 4)
        return self

    def to_bytes(self, **kwargs):
        data = b""
        data += pack_bytes((self["OperandCount"] << 5) + (self["CommandGroup"] << 3) + self["CommandSubGroup"], 1)
        data += pack_bytes((self["DestinationImmediateValueFlag"] << 7) + (self["SourceImmediateValueFlag"] << 6)
                           + (self["reserved1"] << 4) + self["BranchOption"], 1)
        data += pack_bytes((self["reserved2"] << 4) + self["CompareOption"], 1)
        data += pack_bytes((self["reserved3"] << 5) + self["SetOption"], 1)
        data += pack_bytes(self["Destination"], 4)
        data += pack_bytes(self["Source"], 4)
        return data

    def get_command(self):
        if self["CommandGroup"] == CommandGroup.BRANCH:
            if self["CommandSubGroup"] == BranchCommandSubGroup.GOTO:
                if self["BranchOption"] == 0b000:
                    return "Branch_Nop"
                elif self["BranchOption"] == 0b001:
                    return "Branch_GoTo"
                elif self["BranchOption"] == 0b010:
                    return "Branch_Break"
                else:
                    raise ValueError()
            elif self["CommandSubGroup"] == BranchCommandSubGroup.JUMP:
                if self["BranchOption"] == 0b000:
                    return "Branch_JumpObject"
                elif self["BranchOption"] == 0b001:
                    return "Branch_JumpTitle"
                elif self["BranchOption"] == 0b010:
                    return "Branch_CallObject"
                elif self["BranchOption"] == 0b011:
                    return "Branch_CallTitle"
                elif self["BranchOption"] == 0b100:
                    return "Branch_Resume"
                else:
                    raise ValueError()
            elif self["CommandSubGroup"] == BranchCommandSubGroup.PLAY:
                if self["BranchOption"] == 0b000:
                    return "Branch_PlayList"
                elif self["BranchOption"] == 0b001:
                    return "Branch_PlayItem"
                elif self["BranchOption"] == 0b010:
                    return "Branch_PlayMark"
                elif self["BranchOption"] == 0b011:
                    return "Branch_Terminate"
                elif self["BranchOption"] == 0b100:
                    return "Branch_LinkItem"
                elif self["BranchOption"] == 0b101:
                    return "Branch_LinkMark"
                else:
                    raise ValueError()
            else:
                raise ValueError()
        elif self["CommandGroup"] == CommandGroup.CMP:
            if self["CompareOption"] == 0b001:
                return "Compare_BC"
            elif self["CompareOption"] == 0b010:
                return "Compare_EQ"
            elif self["CompareOption"] == 0b011:
                return "Compare_NE"
            elif self["CompareOption"] == 0b100:
                return "Compare_GE"
            elif self["CompareOption"] == 0b101:
                return "Compare_GT"
            elif self["CompareOption"] == 0b110:
                return "Compare_LE"
            elif self["CompareOption"] == 0b111:
                return "Compare_LT"
            else:
                raise ValueError()
        elif self["CommandGroup"] == CommandGroup.SET:
            if self["CommandSubGroup"] == SetCommandSubGroup.SET:
                if self["SetOption"] == 0b00001:
                    return "Set_Move"
                elif self["SetOption"] == 0b00010:
                    return "Set_Swap"
                elif self["SetOption"] == 0b00011:
                    return "Set_Add"
                elif self["SetOption"] == 0b00100:
                    return "Set_Sub"
                elif self["SetOption"] == 0b00101:
                    return "Set_Mul"
                elif self["SetOption"] == 0b00110:
                    return "Set_Div"
                elif self["SetOption"] == 0b00111:
                    return "Set_Mod"
                elif self["SetOption"] == 0b01000:
                    return "Set_Rnd"
                elif self["SetOption"] == 0b01001:
                    return "Set_And"
                elif self["SetOption"] == 0b01010:
                    return "Set_Or"
                elif self["SetOption"] == 0b01011:
                    return "Set_Xor"
                elif self["SetOption"] == 0b01100:
                    return "Set_Bitset"
                elif self["SetOption"] == 0b01101:
                    return "Set_Bitclr"
                elif self["SetOption"] == 0b01110:
                    return "Set_ShiftLeft"
                elif self["SetOption"] == 0b01111:
                    return "Set_ShiftRight"
                else:
                    raise ValueError()
            elif self["CommandSubGroup"] == SetCommandSubGroup.SETSYSTEM:
                if self["SetOption"] == 0b00001:
                    return "Set_SetStream"
                elif self["SetOption"] == 0b00010:
                    return "Set_SetNVTimer"
                elif self["SetOption"] == 0b00011:
                    return "Set_ButtonPage"
                elif self["SetOption"] == 0b00100:
                    return "Set_EnableButton"
                elif self["SetOption"] == 0b00101:
                    return "Set_DisableButton"
                elif self["SetOption"] == 0b00110:
                    return "Set_SetSecondaryStream"
                elif self["SetOption"] == 0b00111:
                    return "Set_PopupOff"
                elif self["SetOption"] == 0b01000:
                    return "Set_StillOn"
                elif self["SetOption"] == 0b01001:
                    return "Set_StillOff"
                else:
                    raise ValueError()
            else:
                raise ValueError()
        else:
            raise ValueError()


class MovieObjectFile:
    def __init__(self, filename=None):
        if not filename:
            self.data = MOBJHeader()
        else:
            self.load(filename)

    def load(self, filename):
        with open(filename, "rb") as f:
            data = f.read()
        self.data = MOBJHeader.from_bytes(data)

    def save(self, destination, overwrite=False):
        self.data.update_constants()
        self.data.update_addresses()
        if os.path.exists(destination) and not overwrite:
            raise FileExistsError()
        os.makedirs(os.path.dirname(destination), exist_ok=True)
        with open(destination, "wb") as f:
            f.write(self.data.to_bytes())



from shinya.common.info_dict import InfoDict


class ExtensionData(InfoDict):
    def __init__(self):
        super().__init__()

    @classmethod
    def from_bytes(cls, data, address=0):
        raise NotImplementedError()

    def check_constraints(self):
        raise NotImplementedError()

    def to_bytes(self, address=0):
        raise NotImplementedError()



from setuptools import setup, find_packages

setup(
    name="shinya",
    version="0.1a1",
    description='A Python package to edit BDMV components like MPLS, CLPI, etc.',
    url='http://github.com/shimamura-hougetsu/shinya',
    author='Shimamura Hougetsu',
    license='MIT',
    packages=find_packages(),
    install_requires=['lxml>=4.6'],
    python_requires=">=3.6"
)
